package org.expeditee.gui;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.MouseInfo;
import java.awt.Point;
import java.awt.RenderingHints;
import java.awt.Toolkit;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.awt.event.WindowStateListener;
import java.util.Collection;

import javax.swing.JFrame;
import javax.swing.RepaintManager;
import javax.swing.SwingUtilities;

import org.expeditee.AbsoluteLayout;
import org.expeditee.actions.Actions;
import org.expeditee.actions.Simple;
import org.expeditee.agents.mail.MailSession;
import org.expeditee.importer.FrameDNDTransferHandler;
import org.expeditee.io.Logger;
import org.expeditee.items.widgets.WidgetCacheManager;
import org.expeditee.stats.StatsLogger;
import org.expeditee.taskmanagement.EntitySaveManager;
import org.expeditee.taskmanagement.SaveStateChangedEvent;
import org.expeditee.taskmanagement.SaveStateChangedEventListener;

/**
 * The Main GUI class, comprises what people will see on the screen.<br>
 * Note: Each Object (Item) is responsible for drawing itself on the screen.<br>
 * Note2: The Frame is registered as a MouseListener and KeyListener, and
 * processes any Events.
 * 
 * @author jdm18
 * 
 */
public class Browser extends JFrame implements ComponentListener,
		WindowListener, WindowStateListener, SaveStateChangedEventListener {

	/**
	 * Default version - just to stop eclipse from complaining about it.
	 */
	private static final long serialVersionUID = 1L;

	// private static final JScrollPane scrollPane = new JScrollPane();

	public static Browser _theBrowser = null;

	// A flag which is set once the application is exiting.
	private boolean _isExiting = false;

	private static boolean _initComplete = false;

	/**
	 * Constructs a new Browser object, then launches it
	 * 
	 * @param args
	 */
	public static void main(String[] args) {

		// Prepare all expeditee and swing data on the AWT event thread.
		SwingUtilities.invokeLater(new Runnable() {
			public void run() {
				// MessageBay.supressMessages(true);

				// MessageBay.supressMessages(false);

				_theBrowser = new Browser();
				_theBrowser.requestFocus();
				FrameMouseActions.MouseX = MouseInfo.getPointerInfo()
						.getLocation().x
						- _theBrowser.getOrigin().x;
				FrameMouseActions.MouseY = MouseInfo.getPointerInfo()
						.getLocation().y
						- _theBrowser.getOrigin().y;
				_initComplete = true;
			}
		});

	}

	public Point getOrigin() {
		return getContentPane().getLocationOnScreen();
	}

	/**
	 * @return
	 * 
	 * True if the application is about to exit. False if not. Not that this is
	 * only set once the window is in its closed state (not closing) or if the
	 * application has explicity being requested to exit.
	 * 
	 * @see Browser#exit()
	 * 
	 */
	public boolean isExisting() {
		return _isExiting;
	}

	private static boolean isInitComplete() {
		return _initComplete;
	}

	public void setSizes(Dimension size) {
		setSize(size);
		setPreferredSize(size);
		Dimension paneSize = getContentPane().getSize();
		FrameGraphics.setMaxSize(paneSize);
	}

	private Browser() {
		// Use the default values initially so we can load the profile frame
		setSizes(new Dimension(UserSettings.InitialWidth,
				UserSettings.InitialHeight));
		// center the frame on the screen
		Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();
		double xpos = screen.getWidth() / 2;
		double ypos = screen.getHeight() / 2;
		setLocation((int) (xpos - (UserSettings.InitialWidth / 2)),
				(int) (ypos - (UserSettings.InitialHeight / 2)));

		addWindowListener(this);
		addWindowStateListener(this);

		DisplayIO.addDisplayIOObserver(WidgetCacheManager.getInstance());
		DisplayIO.addDisplayIOObserver(PopupManager.getInstance());

		UserSettings.Init();
		UserSettings.Username = FrameIO.ConvertToValidFramesetName(System
				.getProperty("user.name"));
		String userName = UserSettings.Username;
		Frame profile = FrameIO.LoadProfile(userName);
		if (profile == null) {
			try {
				profile = FrameIO.CreateNewProfile(userName);
			} catch (Exception e) {
				// TODO tell the user that there was a problem creating the
				// profile frame and close nicely
				e.printStackTrace();
				assert (false);
			}
		}

		// Need to display errors once things have been init otherwise
		// exceptions occur if there are more than four messages neededing to be
		// displayed.
		Collection<String> warningMessages = FrameUtils.ParseProfile(profile);

		// set the layout to absolute layout for widgets
		this.getContentPane().setLayout(new AbsoluteLayout());
		// enable the glasspane-for capturing all mouse events
		this
				.setGlassPane(new MouseEventRouter(getJMenuBar(),
						getContentPane()));
		this.getGlassPane().setVisible(true);
		this.getContentPane().setBackground(Color.white);
		this.getContentPane().setFocusTraversalKeysEnabled(false);

		addComponentListener(this);
		pack();

		// Reset windows to user specified size
		// Must be done after initialising the content pane above!
		setSizes(new Dimension(UserSettings.InitialWidth,
				UserSettings.InitialHeight));

		// See java bug ID 4016934. They say that window closed events are
		// called once the
		// jframe is disposed.
		setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);

		// Expeditee handles its own repainting of AWT/Swing components
		RepaintManager.setCurrentManager(ExpediteeRepaintManager.getInstance());

		// Listen for save status to display during and after runtime
		EntitySaveManager.getInstance().addSaveStateChangedEventListener(this);

		String version = System.getProperty("java.version");

		if (version.startsWith("1.6")) {
			// Set the drag and drop handler
			setTransferHandler(FrameDNDTransferHandler.getInstance());
		} else {
			System.out.println("Upgrade to Java 1.6 to enable drag and drop");
		}

		try {
			warningMessages.addAll(Actions.Init());

			DisplayIO.Init(this);
			// Set visible must be just after DisplayIO.Init for the message box
			// to
			// be the right size
			setVisible(true);

			setupGraphics();

			// required to accept TAB key
			setFocusTraversalKeysEnabled(false);

			// Must be loaded after setupGraphics if images are on the frame
			// Turn off XRay mode and load the first frame
			FrameGraphics.setMode(FrameGraphics.MODE_NORMAL, false);
			FrameUtils.loadFirstFrame(profile);
			DisplayIO.UpdateTitle();

			// I think this can be moved back up to the top of the Go method
			// now...
			// It used to crash the program trying to print error messages up
			// the top
			for (String message : warningMessages)
				MessageBay.warningMessage(message);

			FrameKeyboardActions keyboardListner = new FrameKeyboardActions();
			this.getContentPane().addKeyListener(keyboardListner);
			this.addKeyListener(keyboardListner);

			FrameKeyboardActions.Refresh();
			// setVisible(true);
		} catch (Exception e) {
			e.printStackTrace();
			Logger.Log(e);
		}
	}

	public Graphics2D g;

	private void setupGraphics() {
		if (g != null)
			g.dispose();
		g = (Graphics2D) this.getContentPane().getGraphics();
		assert (g != null);
		g.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING,
				RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
		g.setFont(g.getFont().deriveFont(40f));
		FrameGraphics.setDisplayGraphics(g);
	}

	// private int count = 0;
	@Override
	public void paint(Graphics g) {
		// All this does is make sure the screen is repainted when the browser
		// is moved so that some of the window is off the edge of the display
		// then moved back into view
		super.paint(g);
		FrameGraphics.ForceRepaint();
		// System.out.println("Paint " + count++);
	}

	/**
	 * @inheritDoc
	 */
	public void componentResized(ComponentEvent e) {
		setSizes(this.getSize());
		setupGraphics();
		FrameIO.RefreshCasheImages();
		FrameGraphics.ForceRepaint();
	}

	/**
	 * @inheritDoc
	 */
	public void componentMoved(ComponentEvent e) {
		// FrameGraphics.setMaxSize(this.getSize());
	}

	/**
	 * @inheritDoc
	 */
	public void componentShown(ComponentEvent e) {
	}

	/**
	 * @inheritDoc
	 */
	public void componentHidden(ComponentEvent e) {
	}

	public void windowClosing(WindowEvent e) {
	}

	public void windowClosed(WindowEvent e) {
		exit();
	}

	public void windowOpened(WindowEvent e) {
	}

	public void windowIconified(WindowEvent e) {
	}

	public void windowDeiconified(WindowEvent e) {
	}

	public void windowActivated(WindowEvent e) {
		// System.out.println("Activated");
	}

	public void windowDeactivated(WindowEvent e) {
	}

	public void windowStateChanged(WindowEvent e) {
	}

	public int getDrawingAreaX() {
		// return scrollPane.getLocationOnScreen().x;
		return this.getLocationOnScreen().x;
	}

	public int getDrawingAreaY() {
		// return scrollPane.getLocationOnScreen().y;
		return this.getLocationOnScreen().y;
	}

	public void saveCompleted(SaveStateChangedEvent event) {
		// if (isExisting()) {

		// } else {
		MessageBay.displayMessage("Save finished!", Color.BLUE);
		// }
	}

	public void saveStarted(SaveStateChangedEvent event) {
		// if (isExisting()) {

		// } else {
		String name = event.getEntity().getSaveName();
		if (name == null)
			name = "data";
		MessageBay.displayMessage("Saving " + name + "...", Color.BLUE);
		// }
	}

	/**
	 * Closes the browser and ends the application. Performs saving operations -
	 * halting until saves have completed. Feedback is given to the user while
	 * the application is exiting. Must call on the swing thread.
	 */
	public void exit() {

		// Set exiting flag
		_isExiting = true;

		MessageBay.displayMessage("System exiting...");

		/**
		 * TODO: Prompt the user etc.
		 */

		// TODO: Should we should a popup with a progress bar for user feedback?
		// this would be nice and easy to do.
		// Exit on a dedicated thread so that feedback can be obtained
		new Exiter().start(); // this will exit the application
	}

	/**
	 * The system must exit on a different thread other than the swing thread so
	 * that the save threads can fire save-feedback to the swing thread and thus
	 * provide user feedback on asynchronous save operations.
	 * 
	 * @author Brook Novak
	 * 
	 */
	private class Exiter extends Thread {

		@Override
		public void run() {

			// The final save point for saveable entities
			EntitySaveManager.getInstance().saveAll();
			try {
				EntitySaveManager.getInstance().waitUntilAllSavingFinished();
			} catch (InterruptedException e) {
				e.printStackTrace();
			}

			// The final phase must save on the swing thread since dealing with
			// the expeditee data model
			SwingUtilities.invokeLater(new Runnable() {
				public void run() {

					// Stop any agents or simple programs
					Simple.stop();
					Actions.stopAgent();
					// Wait for them to stop
					try {
						MessageBay
								.displayMessage("Stopping Simple programs..."); // TODO:
						// Only
						// stop
						// if
						// need
						// to...
						while (Simple.isProgramRunning()) {
							Thread.sleep(100); // Brook: What purpose does this
							// serve?
						}
						MessageBay.displayMessage("Stopping Agents..."); // TODO:
						// Only
						// stop
						// if
						// need
						// to...
						while (Actions.isAgentRunning()) {
							Thread.sleep(100); // Brook: What purpose does this
							// serve?
						}
					} catch (Exception e) {

					}

					MessageBay.displayMessage("Saving current frame...");
					FrameIO.SaveFrame(DisplayIO.getCurrentFrame());

					while (FrameIO.DeleteFrameset("messages"))
						;

					MessageBay.displayMessage("Saving stats...");
					StatsLogger.WriteStatsFile();

					if (MailSession.getInstance() != null) {
						MessageBay.displayMessage("Closing ExpMail...");
						MailSession.getInstance().finalise();
					}

					MessageBay.displayMessage("System exited");
					System.exit(0);
				}
			});
		}
	}

	/**
	 * Used to set up the the browser for use in testing.
	 * 
	 * @return
	 */
	public static Browser initializeForTesting() {
		if (Browser._theBrowser == null) {
			MailSession._autoConnect = false;
			Browser.main(null);
			try {
				while (!isInitComplete()) {
					Thread.sleep(10);
				}
			} catch (Exception e) {
			}
		}
		return _theBrowser;
	}

}
