package org.expeditee.gui;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Image;
import java.awt.Polygon;
import java.awt.image.ImageObserver;
import java.awt.image.VolatileImage;
import java.sql.Time;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Stack;

import org.expeditee.actions.Simple;
import org.expeditee.io.Conversion;
import org.expeditee.io.Logger;
import org.expeditee.items.Dot;
import org.expeditee.items.Item;
import org.expeditee.items.ItemAppearence;
import org.expeditee.items.ItemParentStateChangedEvent;
import org.expeditee.items.ItemUtils;
import org.expeditee.items.Line;
import org.expeditee.items.Permission;
import org.expeditee.items.Text;
import org.expeditee.items.XRayable;
import org.expeditee.items.widgets.InteractiveWidget;
import org.expeditee.items.widgets.WidgetCorner;
import org.expeditee.stats.SessionStats;

/**
 * Represents a Expeditee Frame that is displayed on the screen. Also is a
 * registered MouseListener on the Browser, and processes any MouseEvents
 * directly.
 * 
 * @author jdm18
 * 
 */
public class Frame implements ImageObserver {

	private boolean _protectionChanged = false;

	public boolean isReadOnly() {
		return !_frameName.hasPermission(Permission.full)
				&& !_protectionChanged;
	}

	public static Color[] COLOR_WHEEL = { new Color(235, 235, 235),
			new Color(225, 225, 255), new Color(195, 255, 255),
			new Color(225, 255, 225), new Color(255, 255, 195),
			new Color(255, 225, 225), new Color(255, 195, 255), Color.WHITE,
			Color.GRAY, Color.DARK_GRAY, Color.BLACK, null };

	// The various attributes of this Frame
	private String _frameset = null;

	private int _number = -1;

	private int _version = 0;

	private Permission _permission = null;

	private String _owner = null;

	private String _creationDate = null;

	private String _modifiedUser = null;

	private String _modifiedDate = null;

	private String _frozenDate = null;

	// Background color is clear
	private Color _background = null;

	// Foreground color is automatic by default
	private Color _foreground = null;

	public String path;

	private boolean _sorted = true;

	// The items contained in this Frame
	// records whether a change has been made to this Frame (for saving
	// purposes).
	private boolean _change = false;

	private boolean _saved = false;

	// list of deleted items that can be restored
	private Stack<Item> _undo = new Stack<Item>();

	// basically just a list of smaller objects?
	// maybe a hashtable (id -> item?)
	// Note: Needs to be able to be iterated through (for painting)
	private List<Item> _body = new ArrayList<Item>();

	// for drawing purposes
	private List<InteractiveWidget> _iWidgets = new ArrayList<InteractiveWidget>();

	private int _lineCount = 0;

	private int _itemCount = 1;

	// The frameName to display on the screen
	private Text _frameName = null;

	private Map<Overlay, Frame> _overlays = new HashMap<Overlay, Frame>();

	private List<Vector> _vectors = new ArrayList<Vector>();

	private Image _buffer = null;

	private boolean _validBuffer = true;

	private Time _activeTime = new Time(0);

	private Time _darkTime = new Time(0);

	private Collection<Item> _overlayItems = new LinkedHashSet<Item>();

	private Collection<Item> _vectorItems = new LinkedHashSet<Item>();

	/**
	 * Default constructor, nothing is set.
	 */
	public Frame() {
	}

	public Image getBuffer() {
		return _buffer;
	}

	public void setBuffer(Image newBuffer) {
		_buffer = newBuffer;
	}

	public boolean isBufferValid() {
		if (_buffer == null
				|| (_buffer instanceof VolatileImage && ((VolatileImage) _buffer)
						.contentsLost()))
			return false;

		return _validBuffer;
	}

	private void setBufferValid(boolean newValue) {
		_validBuffer = newValue;
	}

	public int getNextItemID() {
		return ++_itemCount;
	}

	public void updateIDs(List<Item> items) {
		for (Item i : items)
			if (!(i instanceof Line))
				i.setID(getNextItemID());
			else
				i.setID(++_lineCount);
	}

	/**
	 * 
	 * @return The interactive widgets that are currently ancored in this frame.
	 *         Hence it exlcudes free-widgets. Returns a copy
	 */
	public List<InteractiveWidget> getInteractiveWidgets() {
		LinkedList<InteractiveWidget> clone = new LinkedList<InteractiveWidget>();
		clone.addAll(this._iWidgets);
		return clone;
	}

	/**
	 * Returns whether this Frame has been changed and required saving to disk.
	 * 
	 * @return True if this Frame has been altered, false otherwise.
	 */
	public boolean hasChanged() {
		// virtual frames are never saved
		if (_number == -1)
			return false;

		return _change;
	}

	/**
	 * Sets whether this Frame should be saved to disk.
	 * 
	 * @param value
	 *            False if this Frame should be saved to disk, False otherwise.
	 */
	public void setChanged(boolean value) {
		// System.out.println(getName() + " " + value);
		boolean oldValue = _change;

		if (oldValue == value)
			return;

		_change = value;

		if (_change) {
			// Notify the frame listeners that the frame has changed
			for (FrameObserver fl : _observers) {
				fl.update();
			}

			setBufferValid(false);
			_saved = false;
		}
	}

	// indicates the frame has changed
	public void change() {
		setChanged(true);
	}

	/**
	 * Returns an ArrayList of all Items currently on the Frame (excludes Items
	 * attached to the cursor).
	 * 
	 * @return The list of Item objects that are on this Frame.
	 */
	public List<Item> getItems(boolean visible) {

		if (!_sorted) {
			Collections.sort(_body);
			_sorted = true;
		}

		List<Item> items = new ArrayList<Item>();

		for (Item i : _body) {
			if (i == null)
				continue;
			if (i.isVisible() || (!visible && !i.isDeleted())) {
				items.add(i);
			}
		}

		return items;
	}

	public List<Item> getItems() {
		return getItems(false);
	}

	/**
	 * @param i
	 *            Item to check if contained in this frame
	 * @return True if this frame contains i.
	 */
	public boolean containsItem(Item i) {
		if (i == null)
			throw new NullPointerException("i");
		return _body.contains(i);
	}

	/**
	 * Returns a list of all the non annotation text items on the frame which
	 * are not the title or frame name or special annotation items.
	 * 
	 * @param includeAnnotations
	 *            true if annotation items without special meaning should be
	 *            included
	 * @param includeLineEnds
	 *            true if text on the end of lines should be included in the
	 *            list
	 * @return the list of body text items.
	 */
	public List<Text> getBodyTextItems(boolean includeAnnotations) {
		List<Text> bodyTextItems = new ArrayList<Text>();
		for (Item i : getItems(true)) {
			// only add up normal body text items
			if ((i instanceof Text)
					&& ((includeAnnotations && !((Text) i)
							.isSpecialAnnotation()) || !i.isAnnotation())
					&& !i.isLineEnd()) {
				bodyTextItems.add((Text) i);
			}
		}
		bodyTextItems.remove(getTitleItem());

		return bodyTextItems;
	}

	public Collection<Item> getNonAnnotationItems(boolean removeTitle) {
		Collection<Item> items = new ArrayList<Item>();
		for (Item i : getItems(true)) {
			// only add up normal body text items
			if (!i.isAnnotation()) {
				items.add(i);
			}
		}
		if (removeTitle) {
			items.remove(getTitleItem());
		}
		return items;
	}

	/**
	 * Gets the last item on the frame that is a non annotation item but is also
	 * text.
	 * 
	 * @return the last non annotation text item.
	 */
	public Item getLastNonAnnotationTextItem() {
		List<Item> items = getItems();

		// find the last non-annotation text item
		for (int i = (items.size() - 1); i >= 0; i--) {
			Item it = items.get(i);

			if (it instanceof Text && !it.isAnnotation()) {
				return (Item) it;
			}
		}
		return null;
	}

	/**
	 * Iterates through the list of items on the frame, and returns one with the
	 * given id if one exists, otherwise returns null.
	 * 
	 * @param id
	 *            The id to search for in the list of items
	 * @return The item on this frame with the given ID, or null if one is not
	 *         found.
	 */
	public Item getItemWithID(int id) {
		for (Item i : _body)
			if (i.getID() == id)
				return i;

		return null;
	}

	/**
	 * Sets this Frame's Title which is displayed in the top left corner.
	 * 
	 * @param title
	 *            The title to assign to this Frame
	 */
	public void setTitle(String title) {
		if (title == null || title.equals(""))
			return;

		boolean oldchange = _change;

		// remove any numbering this title has
		title = title.replaceAll("^\\d*[.] *", "");
		Text frameTitle = getTitleItem();

		if (frameTitle == null) {
			if (UserSettings.TitleTemplate == null) {
				frameTitle = new Text(getNextItemID(), title);
				frameTitle.resetTitlePosition();
			} else {
				frameTitle = UserSettings.TitleTemplate.copy();
				frameTitle.setID(this.getNextItemID());
				frameTitle.setText(title);
			}
			addItem(frameTitle);
		} else {
			// If it begins with a tag remove it

			// Remove the @ symbol if it is there
			// title = ItemUtils.StripTagSymbol(title);
			frameTitle.setText(title);
			// If the @ symbol is followed by numbering or a bullet remove that
			// too
			String autoBulletText = FrameKeyboardActions.getAutoBullet(title);
			if (autoBulletText.length() > 0)
				frameTitle.stripFirstWord();
		}
		// TODO Widgets... check this out
		// Brook: Cannot figure what is going on above... widget annot titles
		// should be stripped always
		if (ItemUtils.startsWithTag(frameTitle, ItemUtils
				.GetTag(ItemUtils.TAG_IWIDGET))) {
			frameTitle.stripFirstWord();
		}

		FrameUtils.Parse(this);

		// do not save if this is the only change
		setChanged(oldchange);
	}

	public Text getTitleItem() {
		List<Item> items = getItems();
		for (Item i : items) {
			if (i instanceof Text && i.getX() < UserSettings.TitlePosition
					&& i.getY() < UserSettings.TitlePosition)
				return (Text) i;
		}

		return null;
	}

	public String getTitle() {
		Text title = getTitleItem();
		if (title == null)
			return getName();

		return title.getFirstLine();
	}

	public Item getNameItem() {
		return _frameName;
	}

	public Text getItemTemplate() {
		return getTemplate(UserSettings.ItemTemplate,
				ItemUtils.TAG_ITEM_TEMPLATE);
	}

	public Text getAnnotationTemplate() {
		Text t = getTemplate(UserSettings.AnnotationTemplate,
				ItemUtils.TAG_ANNOTATION_TEMPLATE);

		if (t == null) {
			t = getItemTemplate();
		}

		return t;
	}

	public Text getStatTemplate() {
		SessionStats.CreatedText();
		Text t = getTemplate(UserSettings.StatTemplate,
				ItemUtils.TAG_STAT_TEMPLATE);

		if (t == null) {
			t = getItemTemplate();
		}

		return t;
	}

	public Item getStatsTextItem(String itemText) {
		return getTextItem(itemText, getStatTemplate());
	}

	public Item getTextItem(String itemText) {
		return getTextItem(itemText, getItemTemplate());
	}

	private Item getTextItem(String itemText, Text template) {
		Text t = template;
		// We dont want the stats to wrap at all
		// t.setMaxWidth(Integer.MAX_VALUE);
		t.setPosition(DisplayIO.getMouseX(), FrameMouseActions.getY());
		// The next line is needed to make sure the item is removed from the
		// frame when picked up
		t.setParent(this);
		t.setText(itemText);
		return t;
	}

	public Text getCodeCommentTemplate() {
		Text t = getTemplate(UserSettings.CodeCommentTemplate,
				ItemUtils.TAG_CODE_COMMENT_TEMPLATE);

		if (t == null) {
			t = getItemTemplate();
		}

		return t;
	}

	/**
	 * Returns any items on this frame that are within the given Shape. Also
	 * returns any Items on overlay frames that are within the Shape.
	 * 
	 * @param shape
	 *            The Shape to search for Items in
	 * @return All Items on this Frame or overlayed Frames for which
	 *         Item.intersects(shape) return true.
	 */
	public Collection<Item> getItemsWithin(Polygon poly) {
		Collection<Item> results = new LinkedHashSet<Item>();
		for (Item i : getVisibleItems()) {
			if (i.intersects(poly)) {
				if (i instanceof XRayable) {
					results.addAll(i.getConnected());
					// Dont add circle centers
					// TODO change this to be isCircle center
				} else if (!i.hasEnclosures()) {
					results.add(i);
				}
			}
		}

		for (Overlay o : _overlays.keySet())
			results.addAll(o.Frame.getItemsWithin(poly));

		return results;
	}

	/**
	 * Sets the name of this Frame to the given String, to be displayed in the
	 * upper right corner.
	 * 
	 * @param name
	 *            The name to use for this Frame.
	 */
	public void setFrameset(String name) {
		_frameset = name;
	}

	public void setName(String framename) {
		int num = Conversion.getFrameNumber(framename);
		String frameset = Conversion.getFramesetName(framename, false);

		setName(frameset, num);
	}

	/**
	 * Sets the frame number of this Frame to the given integer
	 * 
	 * @param number
	 *            The number to set as the frame number
	 */
	public void setFrameNumber(int number) {
		assert (number >= 0);

		if (_number == number)
			return;

		_number = number;
		boolean oldchange = _change;

		int id;

		if (_frameName != null) {
			id = _frameName.getID();
		} else {
			id = -1 * getNextItemID();
		}
		_frameName = new Text(id);
		_frameName.setParent(this);
		_frameName.setText(getFramesetName() + _number);
		_frameName.resetFrameNamePosition();
		setChanged(oldchange);
	}

	/**
	 * Returns the number of this Frame.
	 * 
	 * @return The Frame number of this Frame or -1 if it is not set.
	 */
	public int getNumber() {
		return _number;
	}

	/**
	 * Increments the version of this Frame to the given String.
	 * 
	 * @param version
	 *            The version to use for this Frame.
	 */
	public void setVersion(int version) {
		_version = version;
	}

	/**
	 * Sets the protection of this Frame to the given String.
	 * 
	 * @param protection
	 *            The protection to use for this Frame.
	 */
	public void setPermission(Permission permission) {
		if (_permission != null && _permission.equals(permission))
			_protectionChanged = true;

		_permission = permission;

		if (_body.size() > 0)
			refreshItemPermissions(permission);
	}

	/**
	 * Sets the owner of this Frame to the given String.
	 * 
	 * @param owner
	 *            The owner to use for this Frame.
	 */
	public void setOwner(String owner) {
		_owner = owner;
	}

	/**
	 * Sets the created date of this Frame to the given String.
	 * 
	 * @param date
	 *            The date to use for this Frame.
	 */
	public void setDateCreated(String date) {
		_creationDate = date;
		_modifiedDate = date;
		for (Item i : _body) {
			i.setDateCreated(date);
		}
	}

	public void resetDateCreated() {
		setDateCreated(Logger.EasyDateFormat("ddMMMyyyy:HHmm"));
		setActiveTime(new Time(0));
		setDarkTime(new Time(0));
		setVersion(0);
	}

	/**
	 * Sets the last modifying user of this Frame to the given String.
	 * 
	 * @param user
	 *            The user to set as the last modifying user.
	 */
	public void setLastModifyUser(String user) {
		_modifiedUser = user;
	}

	/**
	 * Sets the last modified date of this Frame to the given String.
	 * 
	 * @param date
	 *            The date to set as the last modified date.
	 */
	public void setLastModifyDate(String date) {
		_modifiedDate = date;
	}

	/**
	 * Sets the last frozen date of this Frame to the given String.
	 * 
	 * @param date
	 *            The date to set as the last frozen date.
	 */
	public void setFrozenDate(String date) {
		_frozenDate = date;
	}

	public void setResort(boolean value) {
		_sorted = !value;
	}

	/**
	 * Adds the given Item to the body of this Frame.
	 * 
	 * @param item
	 *            The Item to add to this Frame.
	 */
	public void addItem(Item item) {
		addItem(item, true);
	}

	public void addItem(Item item, boolean recalculate) {
		if (item == null || item.equals(_frameName) || _body.contains(item))
			return;

		// When an annotation item is anchored the annotation list must be
		// refreshed
		if (item.isAnnotation()) {
			clearAnnotations();
		}

		if (item instanceof Line)
			_lineCount++;

		_itemCount = Math.max(_itemCount, item.getID());

		_body.add(item);
		item.setParent(this);
		item.setFloating(false); // esnure that it is anchored

		item.invalidateCommonTrait(ItemAppearence.Added);

		// If the item is a line end and has constraints with items already
		// on the frame then make sure the constraints hold
		if (item.isLineEnd()) {
			item.setPosition(item.getPosition());
		}

		_sorted = false;

		// item.setMaxWidth(FrameGraphics.getMaxFrameSize().width);
		// add widget items to the list of widgets
		if (item instanceof WidgetCorner) {
			InteractiveWidget iw = ((WidgetCorner) item).getWidgetSource();
			if (!this._iWidgets.contains(iw)) { // A set would have been
				if(FrameMouseActions.isControlDown())
					_iWidgets.add(iw);
				else
					_iWidgets.add(0, iw);
			}
		}

		item.onParentStateChanged(new ItemParentStateChangedEvent(this,
				ItemParentStateChangedEvent.EVENT_TYPE_ADDED));
		if (recalculate && item.recalculateWhenChanged())
			recalculate();

		change();
	}

	public void refreshSize() {
		//assert (size != null);
		boolean bReparse = false;
		for (Item i : getItems()) {
			Float anchorBottom = i.getAnchorBottom();
			Float anchorRight = i.getAnchorRight();
			if (anchorRight != null) {
				i.setAnchorRight(anchorRight);
				if (i.hasVector()) {
					bReparse = true;
				}
			}
			if (anchorBottom != null) {
				i.setAnchorBottom(anchorBottom);
				if (i.hasVector()) {
					bReparse = true;
				}
			}
		}
		
		//Do the anchors on the overlays
		for(Overlay o: getOverlays()){
			o.Frame.refreshSize();
		}
		
		if (bReparse) {
			FrameUtils.Parse(this, false);
		}

		_frameName.resetFrameNamePosition();
	}

	public void addAllItems(Collection<Item> toAdd) {
		for (Item i : toAdd) {
			// If an annotation is being deleted clear the annotation list
			if (i.isAnnotation())
				i.getParentOrCurrentFrame().clearAnnotations();
			// TODO Improve efficiency when addAll is called
			addItem(i);
		}
	}

	public void removeAllItems(Collection<Item> toRemove) {
		for (Item i : toRemove) {
			// If an annotation is being deleted clear the annotation list
			if (i.isAnnotation())
				i.getParentOrCurrentFrame().clearAnnotations();
			removeItem(i);
		}
	}

	public void removeItem(Item item) {
		removeItem(item, true);
	}

	public void removeItem(Item item, boolean recalculate) {
		// If an annotation is being deleted clear the annotation list
		if (item.isAnnotation())
			item.getParentOrCurrentFrame().clearAnnotations();

		if (_body.remove(item)) {
			change();
			// Remove widgets from the widget list
			if (item != null) {
				item.onParentStateChanged(new ItemParentStateChangedEvent(this,
						ItemParentStateChangedEvent.EVENT_TYPE_REMOVED));
				if (item instanceof WidgetCorner) {
					_iWidgets.remove(((WidgetCorner) item).getWidgetSource());
				}
				item.invalidateCommonTrait(ItemAppearence.Removed);
			}
			// TODO Improve efficiency when removeAll is called
			if (recalculate && item.recalculateWhenChanged())
				recalculate();
		}
	}

	/**
	 * Adds the given list of Items to the undo stack. This is the same as
	 * calling addToUndo() for each Item in the list.
	 * 
	 * @param items
	 *            The List of Items to add to the undo stack.
	 */
	public void addAllToUndo(Collection<Item> items) {
		if (items.size() < 1)
			return;

		String id = "" + _undo.size();

		for (Item i : items) {
			i.setTag(id);
			_undo.push(i);
		}
	}

	public void addToUndo(Item item) {
		if (item == null)
			return;

		item.setTag("" + _undo.size());
		_undo.push(item);
	}

	public void undo() {
		Item undo = null;
		boolean bReparse = false;

		if (_undo.size() <= 0)
			return;

		undo = _undo.pop();

		// if the change was to characteristics
		if (undo.isVisible() && _body.contains(undo)) {
			Item old = _body.get(_body.indexOf(undo));
			_body.set(_body.indexOf(old), undo);
			// the item was deleted
		} else {
			List<Item> toRestore = new LinkedList<Item>();
			toRestore.add(undo);

			// remove any connected items at the top of the stack
			while (_undo.size() > 0) {
				Item next = _undo.peek();
				// if this item was connected to one already picked up, remove
				// it from the stack
				if (toRestore.contains(next))
					_undo.pop();
				// else, if this item should be restored (deleted in an enclosed
				// set)
				else if (next.getTag().equals(undo.getTag())) {
					_undo.pop();
					toRestore.add(next);
					// otherwise, we are done
				} else
					break;
			}

			// if these items were deleted from a frame, add them
			addAllItems(toRestore);

			for (Item i : toRestore) {
				bReparse |= i.hasOverlay();
				if (i instanceof Line) {
					Line line = (Line) i;
					line.getStartItem().addLine(line);
					line.getEndItem().addLine(line);
				} else {
					i.setOffset(0, 0);
				}
			}
		}

		change();
		FrameMouseActions.getInstance().refreshHighlights();
		if (bReparse)
			FrameUtils.Parse(this, false);
		FrameGraphics.Repaint();
		ItemUtils.EnclosedCheck(_body);
	}

	/**
	 * Returns the frameset of this Frame
	 * 
	 * @return The name of this Frame's frameset.
	 */
	public String getFramesetName() {
		return _frameset;
	}

	public String getName() {
		return getFramesetName() + _number;
	}

	/**
	 * Returns the format version of this Frame
	 * 
	 * @return The version of this Frame.
	 */
	public int getVersion() {
		return _version;
	}

	public Permission getPermission() {
		return getPermission(Permission.full);
	}

	public Permission getPermission(Permission defaultPermission) {
		if (_permission == null)
			return defaultPermission;

		return _permission;
	}

	public String getOwner() {
		return _owner;
	}

	public String getDateCreated() {
		return _creationDate;
	}

	public String getLastModifyUser() {
		return _modifiedUser;
	}

	public String getLastModifyDate() {
		return _modifiedDate;
	}

	public String getFrozenDate() {
		return _frozenDate;
	}

	public void setBackgroundColor(Color back) {
		_background = back;
		change();

		if (this == DisplayIO.getCurrentFrame()) {
			FrameGraphics.refresh(false);
		}
	}

	public Color getBackgroundColor() {
		return _background;
	}

	public Color getPaintBackgroundColor() {
		// If null... return white
		if (_background == null) {
			return Item.DEFAULT_BACKGROUND;
		}

		return _background;
	}

	public void setForegroundColor(Color front) {
		_foreground = front;
		change();
		// FrameGraphics.Repaint();
	}

	public Color getForegroundColor() {
		return _foreground;
	}

	public Color getPaintForegroundColor() {
		final int GRAY = Color.gray.getBlue();
		final int THRESHOLD = 10;

		if (_foreground == null) {
			Color back = getPaintBackgroundColor();
			if (Math.abs(back.getRed() - GRAY) < THRESHOLD
					&& Math.abs(back.getBlue() - GRAY) < THRESHOLD
					&& Math.abs(back.getGreen() - GRAY) < THRESHOLD)
				return Color.WHITE;

			Color fore = new Color(
					Math.abs(Conversion.RGB_MAX - back.getRed()), Math
							.abs(Conversion.RGB_MAX - back.getGreen()), Math
							.abs(Conversion.RGB_MAX - back.getBlue()));
			return fore;
		}

		return _foreground;
	}

	public String toString() {
		StringBuilder s = new StringBuilder();
		s.append(String.format("Name: %s%d%n", _frameset, _number));
		s.append(String.format("Version: %d%n", _version));
		s.append(String.format("Permission: %s%n", _permission.toString()));
		s.append(String.format("Owner: %s%n", _owner));
		s.append(String.format("Date Created: %s%n", _creationDate));
		s.append(String.format("Last Mod. User: %s%n", _modifiedUser));
		s.append(String.format("Last Mod. Date: %s%n", _modifiedDate));
		s.append(String.format("Items: %d%n", _body.size()));
		return s.toString();
	}

	public Text getTextAbove(Text current) {
		Collection<Text> currentTextItems = FrameUtils.getCurrentTextItems();
		List<Text> toCheck = new ArrayList<Text>();
		if (currentTextItems.contains(current)) {
			toCheck.addAll(currentTextItems);
		} else {
			toCheck.addAll(getTextItems());
		}
		// Make sure the items are sorted
		Collections.sort(toCheck);

		int ind = toCheck.indexOf(current);
		if (ind == -1)
			return null;

		// loop through all items above this one, return the first match
		for (int i = ind - 1; i >= 0; i--) {
			Text check = toCheck.get(i);
			if (FrameUtils.inSameColumn(check, current))
				return check;
		}

		return null;
	}

	/**
	 * Updates any Images that require it from their ImageObserver (Principally
	 * Animated GIFs)
	 */
	public boolean imageUpdate(Image img, int infoflags, int x, int y,
			int width, int height) {
		FrameGraphics.ForceRepaint();

		if (DisplayIO.getCurrentFrame() == this)
			return true;

		return false;
	}

	/**
	 * Gets the text items that are in the same column and below a specified
	 * item. Frame title and name are excluded from the column list.
	 * 
	 * @param from
	 *            The Item to get the column for.
	 */
	public List<Text> getColumn(Item from) {
		// Check that this item is on the current frame
		if (!_body.contains(from))
			return null;

		if (from == null) {
			from = getLastNonAnnotationTextItem();
		}

		if (from == null)
			return null;

		// Get the enclosedItems
		Collection<Text> enclosed = FrameUtils.getCurrentTextItems();
		List<Text> toCheck = null;
		if (enclosed.contains(from)) {
			toCheck = new ArrayList<Text>();
			toCheck.addAll(enclosed);
		} else {
			toCheck = getBodyTextItems(true);
		}

		List<Text> column = new ArrayList<Text>();
		if (toCheck.size() > 0) {

			// Make sure the items are sorted
			Collections.sort(toCheck);

			// Create a list of items consisting of the item 'from' and all the
			// items below it which are also in the same column as it
			int index = toCheck.indexOf(from);

			if (index < 0)
				return column;

			for (int i = index; i < toCheck.size(); i++) {
				Text item = toCheck.get(i);
				if (FrameUtils.inSameColumn(from, item))
					column.add(item);
			}
		}

		return column;
	}

	/**
	 * Adds the given Vector to the list of vector Frames being drawn with this
	 * Frame.
	 * 
	 * @param vector
	 *            The Vector to add
	 * 
	 * @throws NullPointerException
	 *             If overlay is null.
	 */
	protected boolean addVector(Vector toAdd) {
		// make sure we dont add this frame as an overlay of itself
		if (toAdd.Frame == this)
			return false;
		_vectors.add(toAdd);
		// Items must be notified that they have been added or removed from this
		// frame via the vector...
		for (Item i : ItemUtils.CopyItems(toAdd.Frame.getVectorItems(), toAdd)) {
			i.onParentStateChanged(new ItemParentStateChangedEvent(this,
					ItemParentStateChangedEvent.EVENT_TYPE_ADDED_VIA_OVERLAY,
					toAdd.permission));
			_vectorItems.add(i);
			i.invalidateAll();
			i.invalidateFill();
		}
		return true;
	}

	public Collection<Vector> getVectors() {
		Collection<Vector> l = new LinkedList<Vector>();
		l.addAll(_vectors);
		return l;
	}

	public Collection<Overlay> getOverlays() {
		return new LinkedList<Overlay>(_overlays.keySet());
	}

	/**
	 * @return All vectosr seen by this frame (including its vector's vectors).
	 */
	public List<Vector> getVectorsDeep() {
		List<Vector> l = new LinkedList<Vector>();
		getVectorsDeep(l, this, new LinkedList<Frame>());
		return l;
	}

	private boolean getVectorsDeep(List<Vector> vectors, Frame vector,
			List<Frame> seenVectors) {

		if (seenVectors.contains(vector))
			return false;

		seenVectors.add(vector);

		for (Vector o : vector.getVectors()) {
			if (getVectorsDeep(vectors, o.Frame, seenVectors)) {
				vectors.add(o);
			}
		}

		return true;
	}

	// private boolean getOverlaysDeep(List<Overlay> overlays, Frame overlay,
	// List<Frame> seenOverlays) {
	//
	// if (seenOverlays.contains(overlay))
	// return false;
	//
	// seenOverlays.add(overlay);
	//
	// for (Overlay o : overlay.getOverlays()) {
	// if (getOverlaysDeep(overlays, o.Frame, seenOverlays)) {
	// overlays.add(o);
	// }
	// }
	//
	// return true;
	// }

	/**
	 * Gets the overlay on this frame which owns the given item.
	 * 
	 * @param item
	 *            The item - must not be null.
	 * @return The overlay that contains the itm. Null if no overlay owns the
	 *         item.
	 */
	public Overlay getOverlayOwner(Item item) {
		if (item == null)
			throw new NullPointerException("item");

		for (Overlay l : getOverlays()) {
			if (item.getParent() == l.Frame)
				return l;
		}

		// TODO return the correct vector... not just the first vector matching
		// the vectorFrame
		for (Vector v : getVectors()) {
			if (item.getParent() == v.Frame)
				return v;
		}

		return null;
	}

	public void clearVectors() {
		_vectors.clear();

		for (Item i : _vectorItems) { // TODO: Rethink where this should live
			i.invalidateAll();
			i.invalidateFill();
		}
		_vectorItems.clear();

	}

	protected boolean removeVector(Vector toRemove) {
		if (!_vectors.remove(toRemove))
			return false;
		for (Item i : toRemove.Frame.getVectorItems()) {
			i.invalidateAll();
			i.invalidateFill();
			_vectorItems.remove(i);
			i.onParentStateChanged(new ItemParentStateChangedEvent(this,
					ItemParentStateChangedEvent.EVENT_TYPE_REMOVED_VIA_OVERLAY,
					toRemove.permission));

		}
		return true;
	}

	public void clearOverlays() {
		for (Overlay o : _overlays.keySet()) {
			for (Item i : o.Frame.getItems()) {
				i
						.onParentStateChanged(new ItemParentStateChangedEvent(
								this,
								ItemParentStateChangedEvent.EVENT_TYPE_REMOVED_VIA_OVERLAY,
								o.permission));
			}
		}
		_overlayItems.clear();
		_overlays.clear();
		assert (_overlays.isEmpty());
	}

	protected boolean removeOverlay(Frame f) {
		for (Overlay o : _overlays.keySet()) {
			if (o.Frame == f) {
				_overlays.remove(o);
				for (Item i : f.getItems()) {
					_overlayItems.remove(i);
					i
							.onParentStateChanged(new ItemParentStateChangedEvent(
									this,
									ItemParentStateChangedEvent.EVENT_TYPE_REMOVED_VIA_OVERLAY,
									o.permission));
				}
				return true;
			}
		}
		return false;
	}

	public void addAllVectors(List<Vector> vectors) {
		for (Vector v : vectors) {
			addVector(v);
		}
	}

	public void addAllOverlays(Collection<Overlay> overlays) {
		for (Overlay o : overlays) {
			addOverlay(o);
		}
	}

	protected boolean addOverlay(Overlay toAdd) {
		// make sure we dont add this frame as an overlay of itself
		if (toAdd.Frame == this)
			return false;
		// Dont add the overlay if there is already one for this frame
		if (_overlays.values().contains(toAdd.Frame))
			return false;
		// Add the overlay to the map of overlays on this frame
		_overlays.put(toAdd, toAdd.Frame);
		// Add all the overlays from the overlay frame to this frame
		for (Overlay o : toAdd.Frame.getOverlays())
			addOverlay(o);

		// Now add the items for this overlay
		Permission permission = Permission.min(toAdd.Frame.getPermission(),
				toAdd.permission);

		// Items must be notified that they have been added or removed from this
		// frame via the overlay...
		for (Item i : toAdd.Frame.getVisibleItems()) {
			i.onParentStateChanged(new ItemParentStateChangedEvent(this,
					ItemParentStateChangedEvent.EVENT_TYPE_ADDED_VIA_OVERLAY,
					permission));
			//i.setPermission(permission);
			_overlayItems.add(i);
		}
		return true;
	}

	@Override
	public boolean equals(Object o) {
		if (o instanceof String) {
			return (String.CASE_INSENSITIVE_ORDER
					.compare((String) o, getName()) == 0);
		}

		if (o instanceof Frame) {
			return getName().equals(((Frame) o).getName());
		}

		return super.equals(o);
	}

	/**
	 * Merge one frames contents into another.
	 * 
	 * @param toMergeWith
	 */
	private void merge(Frame toMergeWith) {
		if (toMergeWith == null)
			return;

		List<Item> copies = ItemUtils.CopyItems(toMergeWith.getItems());
		copies.remove(toMergeWith.getNameItem());

		for (Item i : copies) {
			if (i.getID() >= 0) {
				i.setID(this.getNextItemID());
				addItem(i);
			}
		}
	}

	/**
	 * This method is for merging frames or setting frame attributes via
	 * injecting a text item into the frameName item.
	 * 
	 * @param toMerge
	 * @return the items that cant be merged
	 */
	public List<Item> merge(List<Item> toMerge) {
		ArrayList<Item> remain = new ArrayList<Item>(0);

		for (Item i : toMerge) {
			if (!(i instanceof Text))
				remain.add(i);
			else {
				if (!AttributeUtils.setAttribute(this, (Text) i)) {
					if (i.getLink() != null)
						merge(FrameIO.LoadFrame(i.getAbsoluteLink()));
					else if (FrameIO
							.isValidFrameName(((Text) i).getFirstLine())) {
						// If we get hear we are merging frames
						merge(FrameIO.LoadFrame(((Text) i).getFirstLine()));
					}
				}
			}
		}

		return remain;
	}

	/**
	 * Removes all non-title non-annotation items from this Frame. All removed
	 * items are added to the backup-stack.
	 */
	public void clear(boolean keepAnnotations) {
		List<Item> newBody = new ArrayList<Item>(0);
		Item title = getTitleItem();
		if (title != null) {
			newBody.add(title);
			_body.remove(title);
		}
		if (keepAnnotations) {
			for (Item i : _body) {
				if (i.isAnnotation())
					newBody.add(i);
			}
		}
		_body.removeAll(newBody);
		addAllToUndo(_body);
		_body = newBody;
		change();

		if (!keepAnnotations && _annotations != null)
			_annotations.clear();
	}

	/**
	 * Creates a new text item with the given text.
	 * 
	 * @param text
	 * @return
	 */
	public Text createNewText(String text) {
		Text t = createBlankText(text);
		t.setText(text);
		return t;
	}

	/**
	 * Creates a new Text Item with no text. The newly created Item is a copy
	 * the ItemTemplate if one is present, and inherits all the attributes of
	 * the Template
	 * 
	 * @return The newly created Text Item
	 */
	public Text createBlankText(String templateType) {
		SessionStats.CreatedText();
		Text t;
		if (templateType.length() == 0)
			t = getItemTemplate().copy();
		else
			t = getItemTemplate(templateType.charAt(0));

		// reset attributes
		t.setID(getNextItemID());
		t.setPosition(DisplayIO.getMouseX(), FrameMouseActions.getY());
		t.setText("");
		t.setParent(this);

		// Set the width if the template doesnt have a width
		// Make it the width of the page
		// t.setMaxWidth(FrameGraphics.getMaxFrameSize().width);
		// if (t.getWidth() <= 0) {
		t.setRightMargin(FrameGraphics.getMaxFrameSize().width);
		// }
		addItem(t);
		return t;
	}

	public Item createDot() {
		Item dot = new Dot(DisplayIO.getMouseX(), FrameMouseActions.getY(),
				getNextItemID());

		Item template = getTemplate(UserSettings.DotTemplate,
				ItemUtils.TAG_DOT_TEMPLATE);
		float thickness = template.getThickness();
		if (thickness > 0)
			dot.setThickness(template.getThickness());
		if (template.getLinePattern() != null)
			dot.setLinePattern(template.getLinePattern());
		dot.setColor(template.getColor());
		dot.setFillColor(template.getFillColor());
		// reset attributes
		dot.setParent(this);
		return dot;
	}

	private Text getTemplate(Text defaultTemplate, int templateTag) {
		Text t = null;

		// check for an updated template...
		for (Item i : this.getItems()) {
			if (ItemUtils.startsWithTag(i, templateTag)) {
				t = (Text) i;
				break;
			}
		}

		if (t == null) {
			if (defaultTemplate == null) {
				return null;
			}
			t = defaultTemplate;
		}

		// If the item is linked apply any attribute pairs on the child frame
		String link = t.getAbsoluteLink();
		// need to get link first because copy doesnt copy the link
		t = t.copy();
		if (link != null) {
			t.setLink(null);
			Frame childFrame = FrameIO.LoadFrame(link);
			if (childFrame != null) {
				// read in attribute value pairs
				for (Text attribute : childFrame.getBodyTextItems(false)) {
					AttributeUtils.setAttribute(t, attribute);
				}
			}
		}
		return t;
	}

	private Text getItemTemplate(char firstChar) {
		switch (firstChar) {
		case '@':
			return getAnnotationTemplate();
		case '/':
		case '#':
			return getCodeCommentTemplate();
		default:
			return getItemTemplate();
		}
	}

	public Text createNewText() {
		return createNewText("");
	}

	public Text addText(int x, int y, String text, String action) {
		Text t = createNewText(text);
		t.setPosition(x, y);
		t.addAction(action);
		return t;
	}

	public Item addText(int x, int y, String text, String action, String link) {
		Item t = addText(x, y, text, action);
		t.setLink(link);
		return t;
	}

	public Item addDot(int x, int y) {
		Item d = new Dot(x, y, getNextItemID());
		addItem(d);
		return d;
	}

	public boolean isSaved() {
		return _saved;
	}

	public void setSaved() {
		// System.out.println(getName() + " saved");
		_saved = true;
		_change = false;
	}

	public static boolean rubberbandingLine() {
		return FreeItems.getInstance().size() == 2
				&& (FreeItems.getInstance().get(0) instanceof Line || FreeItems
						.getInstance().get(1) instanceof Line);
	}

	/**
	 * Tests if an item is a non title, non frame name, non special annotation
	 * text item.
	 * 
	 * @param it
	 *            the item to be tested
	 * @return true if the item is a normal text item
	 */
	public boolean isNormalTextItem(Item it) {
		if (it instanceof Text && it != getTitleItem() && it != _frameName
				&& !((Text) it).isSpecialAnnotation()) {
			return true;
		}

		return false;
	}

	/**
	 * Moves the mouse to the end of the text item with a specified index.
	 * 
	 * @param index
	 */
	public boolean moveMouseToTextItem(int index) {
		List<Item> items = getItems();
		int itemsFound = 0;
		for (int i = 0; i < items.size(); i++) {
			Item it = items.get(i);
			if (isNormalTextItem(it))
				itemsFound++;
			if (itemsFound > index) {
				DisplayIO.setCursorPosition(((Text) it)
						.getParagraphEndPosition().x, it.getY());
				DisplayIO.resetCursorOffset();
				FrameGraphics.Repaint();
				return true;
			}
		}

		return false;
	}

	/*
	 * public boolean moveMouseToNextTextItem(int index) { List<Item> items =
	 * getItems(); int itemsFound = 0; for (int i = 0; i < items.size(); i++) {
	 * Item it = items.get(i); if ( isNormalTextItem(it)) itemsFound++; if
	 * (itemsFound > index) {
	 * DisplayIO.setCursorPosition(((Text)it).getEndParagraphPosition().x,
	 * it.getY()); DisplayIO.resetCursorOffset(); FrameGraphics.Repaint();
	 * return true; } }
	 * 
	 * return false; }
	 */

	/**
	 * Searches for an annotation item called start to be used as the default
	 * cursor location when TDFC occurs.
	 */
	public boolean moveMouseToDefaultLocation() {
		List<Item> items = getItems();

		for (Item it : items) {
			if (it instanceof Text) {
				Text t = (Text) it;
				if (t.getText().toLowerCase().startsWith("@start")
						|| t.getText().toLowerCase().equals("@start:")) {
					// Used to allow users the option of putting an initial
					// bullet after the @start
					// This was replaced by width
					// t.stripFirstWord();
					t.setText("");

					if (t.getText().equals(""))
						DisplayIO.getCurrentFrame().removeItem(t);
					if (!FreeItems.itemsAttachedToCursor()) {
						DisplayIO.setCursorPosition(((Text) it)
								.getParagraphEndPosition());
						DisplayIO.resetCursorOffset();
					}
					FrameGraphics.Repaint();
					return true;
				}
			}
		}

		return false;
	}

	/**
	 * Gets the file name that actions should use to export files created by
	 * running actions from this frame.
	 * 
	 * @return the fileName if the frame contains an '@file' tag. Returns the
	 *         name of the frame if the tag isnt on the frame.
	 */
	public String getExportFileName() {
		Text tag = ItemUtils.FindTag(getItems(), "@file:");
		if (tag != null) {
			String file = ItemUtils.StripTag(tag.getFirstLine(), "@file");
			file = file.trim();
			if (file.length() > 0)
				return file;
		}

		return _frameName.getText();
	}

	public void toggleBackgroundColor() {
		setBackgroundColor(ColorUtils.getNextColor(_background,
				Frame.COLOR_WHEEL, null));
	}

	public void setName(String frameset, int i) {
		setFrameset(frameset);
		setFrameNumber(i);
	}

	/**
	 * Sets the item permissions to match the protection for the frame.
	 * 
	 */
	public void refreshItemPermissions(Permission maxPermission) {
		Permission permission = Permission.min(maxPermission, getPermission());
		_frameName.setPermission(permission);

		switch (permission) {
		case none:
			_frameName.setBackgroundColor(new Color(255, 220, 220));
			break;
		case followLinks:
			_frameName.setBackgroundColor(new Color(255, 230, 135));
			break;
		case copy:
			_frameName.setBackgroundColor(new Color(255, 255, 155));
			break;
		case createFrames:
			_frameName.setBackgroundColor(new Color(220, 255, 220));
			break;
		case full:
			_frameName.setBackgroundColor(null);
			break;
		default:
			assert (false);
			break;
		}

		for (Overlay o : getOverlays())
			o.Frame.refreshItemPermissions(o.permission);
		
		//Only update the permissions if we have to
		if (_body.size() > 0 && permission.equals(_body.get(0)._permission))
			return;

		for (Item i : _body) {
			i.setPermission(permission);
		}
		

	}

	public boolean isTestFrame() {
		Text title = getTitleItem();
		if (title == null)
			return false;
		String action = title.getFirstAction();
		if (action == null)
			return false;
		action = action.toLowerCase();
		return action.startsWith(Simple.RUN_FRAME_ACTION)
				|| action.startsWith(Simple.DEBUG_FRAME_ACTION);
	}

	public void setActiveTime(String activeTime) {
		try {
			_activeTime = new Time(Time.valueOf(activeTime).getTime() + 12 * 60
					* 60 * 1000);
		} catch (Exception e) {
			_activeTime = new Time(0);
		}
	}

	public void setActiveTime(Time activeTime) {
		_activeTime = activeTime;
	}

	public void setDarkTime(Time darkTime) {
		_darkTime = darkTime;
	}

	public void setDarkTime(String darkTime) {
		try {
			_darkTime = new Time(Time.valueOf(darkTime).getTime() + 12 * 60
					* 60 * 1000);
		} catch (Exception e) {
			_darkTime = new Time(0);
		}
	}

	/**
	 * Returns null if their is no backup frame or if it is invalid.
	 * 
	 * @return the backup frame for this frame
	 */
	public Frame getBackupFrame() {
		Text backupTag = _annotations.get("old");
		if (backupTag == null)
			return null;
		// TODO want another way to deal with updating of annotations items
		// without F12 refresh
		// Reparse the frame if annotation item has been modified
		String[] processedText = backupTag.getProcessedText();
		if (processedText == null) {
			// Reparse the frame if this item has not yet been parsed
			FrameUtils.Parse(this);
			return getBackupFrame();
		}
		// Now return the name of the backed up frame
		String link = backupTag.getAbsoluteLink();
		if (link == null || link.equalsIgnoreCase(getName()))
			return null;

		Frame backup = FrameIO.LoadFrame(link);
		return backup;
	}

	public Time getDarkTime() {
		return _darkTime;
	}

	public Time getActiveTime() {
		return _activeTime;
	}

	/**
	 * Gets the number of backed up versions of this frame are saved plus 1 for
	 * this frame.
	 * 
	 * @return the number of frames in the backed up comet
	 */
	public int getCometLength() {
		Frame backup = getBackupFrame();
		return 1 + (backup == null ? 0 : backup.getCometLength());
	}

	public void addAnnotation(Text item) {
		if (_annotations == null) {
			_annotations = new HashMap<String, Text>();
		}
		// Check if this item has already been processed
		String[] tokens = item.getProcessedText();
		if (tokens != null) {
			if (tokens.length > 0) {
				_annotations.put(tokens[0], item);
			}
			return;
		}

		String text = item.getText().trim();
		assert (text.charAt(0) == '@');
		// Ignore annotations with spaces after the tag symbol
		if (text.length() < 2 || !Character.isLetter(text.charAt(1))) {
			item.setProcessedText(new String[0]);
			return;
		}
		// The separator char must come before the first non letter otherwise we
		// ignore the annotation item
		for (int i = 2; i < text.length(); i++) {
			char ch = text.charAt(i);
			if (!Character.isLetterOrDigit(ch)) {
				// Must have an attribute value pair
				if (ch == AttributeValuePair.SEPARATOR_CHAR) {
					// Get the attribute
					String attribute = text.substring(1, i).toLowerCase();
					String value = "";
					if (text.length() > 1 + i) {
						value = text.substring(i + 1).trim();
					}
					item.setProcessedText(new String[] { attribute, value });
					_annotations.put(attribute, item);
					return;
				} else {
					item.setProcessedText(new String[0]);
					return;
				}
			}
		}
		// If it was nothing but letters and digits save the tag
		String lowerCaseText = text.substring(1).toLowerCase();
		item.setProcessedText(new String[] { lowerCaseText });
		_annotations.put(lowerCaseText, item);
	}

	public boolean hasAnnotation(String annotation) {
		if (_annotations == null)
			refreshAnnotationList();
		return _annotations.containsKey(annotation.toLowerCase());
	}

	/**
	 * Returns the annotation value in full case.
	 * 
	 * @param annotation
	 *            the annotation to retrieve the value of.
	 * @return the annotation item value in full case or null if the annotation
	 *         is not on the frame or has no value.
	 */
	public String getAnnotationValue(String annotation) {
		if (_annotations == null)
			refreshAnnotationList();

		Text text = _annotations.get(annotation.toLowerCase());
		if (text == null)
			return null;

		String[] tokens = text.getProcessedText();
		if (tokens != null && tokens.length > 1)
			return tokens[1];
		return null;
	}

	Map<String, Text> _annotations = null;

	private Collection<FrameObserver> _observers = new HashSet<FrameObserver>();

	public void clearAnnotations() {
		_annotations = null;
	}

	public List<Item> getVisibleItems() {
		return getItems(true);
	}

	private void refreshAnnotationList() {
		if (_annotations == null)
			_annotations = new HashMap<String, Text>();
		else
			_annotations.clear();
		for (Text text : getTextItems()) {
			if (text.isAnnotation()) {
				addAnnotation(text);
			}
		}
	}

	public Collection<Text> getAnnotationItems() {
		if (_annotations == null) {
			refreshAnnotationList();
		}
		return _annotations.values();
	}

	/**
	 * Gets a list of items to be saved to file by text file writers.
	 * 
	 * @return the list of items to be saved to a text file
	 */
	public List<Item> getItemsToSave() {

		if (!_sorted) {
			Collections.sort(_body);
			_sorted = true;
		}

		// iWidgets are handled specially since 8 items are written as one
		Collection<InteractiveWidget> seenWidgets = new LinkedHashSet<InteractiveWidget>();

		List<Item> toSave = new ArrayList<Item>();

		for (Item i : _body) {
			if (i == null || i.dontSave()) {
				continue;
			}

			// Ensure only one of the WidgetCorners represent a single widget
			if (i instanceof WidgetCorner) {
				InteractiveWidget iw = ((WidgetCorner) i).getWidgetSource();
				if (seenWidgets.contains(iw))
					continue;
				seenWidgets.add(iw);
				toSave.add(iw.getSource());
			} else if (i instanceof XRayable) {
				XRayable x = (XRayable) i;
				toSave.addAll(x.getItemsToSave());
			}// Circle centers are items with attached enclosures
			else if (i.hasEnclosures()) {
				continue;
			} else {
				toSave.add(i);
			}
		}
		return toSave;
	}

	public Collection<Item> getOverlayItems() {
		return _overlayItems;
	}

	/**
	 * Returns true if this frame has and overlays for the specified frame.
	 * 
	 * @param frame
	 * @return
	 */
	public boolean hasOverlay(Frame frame) {
		return _overlays.containsValue(frame);
	}

	public Collection<Item> getAllItems() {
		Collection<Item> allItems = new LinkedHashSet<Item>(_body);
		allItems.addAll(_overlayItems);
		allItems.addAll(_vectorItems);
		return allItems;
	}

	public Collection<Item> getVectorItems() {
		Collection<Item> vectorItems = new LinkedHashSet<Item>(_vectorItems);
		vectorItems.addAll(getNonAnnotationItems(false));
		return vectorItems;
	}

	/**
	 * Gets a list of all the text items on the frame.
	 * 
	 * @return
	 */
	public Collection<Text> getTextItems() {
		Collection<Text> textItems = new ArrayList<Text>();
		for (Item i : getItems(true)) {
			// only add up normal body text items
			if ((i instanceof Text)) {
				textItems.add((Text) i);
			}
		}
		return textItems;
	}

	public Text getAnnotation(String annotation) {
		if (_annotations == null)
			refreshAnnotationList();

		return _annotations.get(annotation.toLowerCase());
	}

	public void recalculate() {
		for (Item i : getItems()) {
			if (i.hasFormula() && !i.isAnnotation()) {
				i.calculate(i.getFormula());
			}
		}
	}

	public void removeObserver(FrameObserver observer) {
		_observers.remove(observer);
	}

	public void addObserver(FrameObserver observer) {
		_observers.add(observer);
	}

	public void clearObservers() {
		for (FrameObserver fl : _observers) {
			fl.removeSubject(this);
		}
		// The frame listener will call the frames removeListener method
		assert (_observers.size() == 0);
	}

	public Collection<Text> getNonAnnotationText(boolean removeTitle) {
		Collection<Text> items = new LinkedHashSet<Text>();
		for (Item i : getItems(true)) {
			// only add up normal body text items
			if (i instanceof Text && !i.isAnnotation()) {
				items.add((Text) i);
			}
		}
		if (removeTitle) {
			items.remove(getTitleItem());
		}
		return items;
	}

	public void dispose() {
		clearObservers();
		for (Item i : _body) {
			i.dispose();
		}
		_frameName.dispose();
		_body = null;
		_frameName = null;
	}
}
